/*
	$Id: Class.java 321 2007-05-25 14:23:03Z syl $

	Copyright (C) 2006 Syllabub
	Maroontress Fast Software.
*/

package com.maroontress.sanctuary;

import java.io.File;
import java.io.PrintWriter;
import java.util.Arrays;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import org.w3c.dom.Element;

/**
   饹Ǥ
*/
public class Class extends Group {

    /** 󥯥롼ɤΥå */
    private Includes includes;

    /** ݡȤΥå */
    private Imports imports;

    /** ݡȤΥå */
    private Exports exports;

    /** ȥåץ٥ι¤ΡΡ󷿤Υå */
    private ToplevelCompounds compounds;

    /** ᥽åɤΥå */
    private RegularMethods methods;

    /** ֥ȥ饯ȥ᥽åɤΥå */
    private AbstractMethods abstractMethods;

    /** ޥΥå */
    private Macros macros;

    /** Υå */
    private Typedefs typedefs;

    /** ХѿΥå */
    private GlobalVariables globalVariables;

    /** Υ饹˽°빽¤ */
    private LinkedList<Compound> structs;

    /** Υ饹˽°붦 */
    private LinkedList<Compound> unions;

    /** Υ饹˽° */
    private LinkedList<Compound> enums;

    /** ݡȤ륿Υꥹ */
    private HashSet<String> tags;

    /** ¤ΡΤΥɽ */
    private static final String TAG_REGEX = "(struct|union) +[_A-Za-z]\\w*";

    /** ¤ΡΤΥɽΥѥѤߥѥ */
    private static Pattern tagPattern;

    static {
	tagPattern = Pattern.compile(TAG_REGEX);
    }

    /**
       饹ޤ

       @param elem class
       @param filename XMLΥե̾
    */
    public Class(final Element elem, final String filename) {
	super(elem, filename);

	structs = new LinkedList<Compound>();
	unions = new LinkedList<Compound>();
	enums = new LinkedList<Compound>();
	tags = new HashSet<String>();

	ChildrenFilter c = getChildrenFilter();
        includes = new Includes(c.getDirectChildrenByTagName("include"));

	exports = new Exports(this, c.getDirectChildrenByTagName("export"));
	methods = new RegularMethods(
	    this, c.getDirectChildrenByTagName("method"));
	compounds = new ToplevelCompounds(
	    this, c.getDirectChildrenByTagName("struct|union|enum"));
	abstractMethods = new AbstractMethods(
	    this, c.getDirectChildrenByTagName("abstractmethod"));
	macros = new Macros(this, c.getDirectChildrenByTagName("define"));
	typedefs = new Typedefs(this, c.getDirectChildrenByTagName("typedef"));
	globalVariables = new GlobalVariables(
	    this, c.getDirectChildrenByTagName("globalvariable"));
	// ĹimportǤθФΤᡢImportsϺǸ
	imports = new Imports(this, c.getDirectChildrenByTagName("import"));
    }

    /**
       إåե˥ݡȤ륿Ϥޤ

       @param wr 饤
    */
    private void printImportedTags(final PrintWriter wr) {
	if (tags.size() == 0) {
	    return;
	}
	wr.println("");
	String[] a = tags.toArray(new String[0]);
	Arrays.sort(a);
	for (String s : a) {
	    wr.println(s + ";");
	}
    }

    /**
       {@inheritDoc}
    */
    @Override public void printCode() {
	String packageName = getPackageName();
	String name = getName();
	String dirName = packageName.replace('_', File.separatorChar);
	PrintWriter wr = createPrintWriter(dirName, name + ".h");
	String includeMacro = packageName + "_" + name + "_H";
	String importMacro = packageName + "_" + name + "_IMPORT";

	wr.println("#ifndef " + includeMacro);
	wr.println("#define " + includeMacro);

	includes.printCode(wr);
	macros.printCode(wr);
	printImportedTags(wr);
	typedefs.printCode(wr);
	compounds.printCode(wr);
	methods.printCode(wr);
	globalVariables.printCode(wr);

	if (isNamespaceExpanded()) {
	    wr.println("");
	    wr.println("#ifdef " + importMacro);
	    printImportMacro(wr);
	    wr.println("#endif /* " + importMacro + " */");
	}

	wr.println("");
	wr.println("#endif /* !" + includeMacro + " */");
	wr.close();
    }

    /**
       {@inheritDoc}
    */
    @Override protected String getType() {
	return "class";
    }

    /**
       ѥɤHTMLǽϤޤ

       @param wr 饤
       @param label "Struct" ޤ "Union" ޤ "Enum"
       @param s ѥɤΥꥹ
    */
    private void printCompoundSummary(final PrintWriter wr, final String label,
				      final List<Compound> s) {
	if (s.size() == 0) {
	    return;
	}

	wr.println(""
+ "    <h2>" + label + " Summary</h2>");
	for (Compound c : s) {
	    c.printCompoundSummary(wr);
	}
    }

    /**
       ѥɤξܺ٤HTMLǽϤޤ

       @param wr 饤
       @param label "Struct" ޤ "Union" ޤ "Enum"
       @param s ѥɤΥꥹ
    */
    private void printCompoundDetail(final PrintWriter wr, final String label,
				     final List<Compound> s) {
	if (s.size() == 0) {
	    return;
	}
	wr.println(""
+ "    <h2>" + label + " Detail</h2>");
	for (Compound c : s) {
	    c.printCompoundDetail(wr);
	}
    }

    /**
       {@inheritDoc}
    */
    @Override protected void printHtmlUsage(final PrintWriter wr) {
	String s = getPackageName() + "_" + getName();
	String h = s.replace('_', File.separatorChar) + ".h";
	if (isNamespaceExpanded()) {
	    wr.println("#define " + s + "_IMPORT<br>");
	}
	wr.println("#include &lt;" + h + "&gt;<br>");
    }

    /**
       {@inheritDoc}
    */
    @Override protected void printHtml(final PrintWriter wr) {
	macros.printSummary(wr);
	typedefs.printSummary(wr);
	printCompoundSummary(wr, "Struct", structs);
	printCompoundSummary(wr, "Union", unions);
	printCompoundSummary(wr, "Enum", enums);
	methods.printSummary(wr);
	abstractMethods.printSummary(wr);
	globalVariables.printSummary(wr);

	macros.printDetail(wr);
	typedefs.printDetail(wr);
	printCompoundDetail(wr, "Struct", structs);
	printCompoundDetail(wr, "Union", unions);
	printCompoundDetail(wr, "Enum", enums);
	methods.printDetail(wr);
	abstractMethods.printDetail(wr);
	globalVariables.printDetail(wr);
    }

    /**
       ѥɤξܺ٤TeXǽϤޤ

       @param wr 饤
       @param label "Struct" ޤ "Union" ޤ "Enum"
       @param s ѥɤΥꥹ
    */
    private void printTexCompoundDetail(final PrintWriter wr,
					final String label,
					final List<Compound> s) {
	if (s.size() == 0) {
	    return;
	}
	wr.println("\\S" + label.toLowerCase());
	for (Compound c : s) {
	    c.printTexCompoundDetail(wr);
	}
    }

    /**
       {@inheritDoc}
    */
    @Override protected void printTexUsage(final PrintWriter wr) {
	String s = getPackageName() + "_" + getName();
	String h = s.replace('_', File.separatorChar) + ".h";
	wr.println("\\nopagebreak");
	wr.println("\\begin{Ssyntax}%");
	if (isNamespaceExpanded()) {
	    wr.println(Tex.safe("#define " + s + "_IMPORT") + "\\\\");
	}
	wr.println(Tex.safe("#include <" + h + ">") + "\\\\");
	wr.println("\\end{Ssyntax}");
    }

    /**
       {@inheritDoc}
    */
    @Override protected void printTex(final PrintWriter wr) {
	macros.printTexDetail(wr);
	typedefs.printTexDetail(wr);
	printTexCompoundDetail(wr, "Struct", structs);
	printTexCompoundDetail(wr, "Union", unions);
	printTexCompoundDetail(wr, "Enum", enums);
	methods.printTexDetail(wr);
	abstractMethods.printTexDetail(wr);
	globalVariables.printTexDetail(wr);
    }

    /**
       {@inheritDoc}
    */
    @Override public void printStub() {
	printStubHeader();
	printStubCode();
    }

    /**
       ֤Υإåեޤ
    */
    private void printStubHeader() {
	String packageName = "stub_" + getPackageName();
	String name = getName();
	String dirName = packageName.replace('_', File.separatorChar);
	PrintWriter wr = createPrintWriter(dirName, name + ".h");
	String includeMacro = packageName + "_" + name + "_H";

	wr.println("#ifndef " + includeMacro);
	wr.println("#define " + includeMacro);

	String s = getPackageName() + "_" + getName();
	s = s.replace('_', File.separatorChar);

	wr.println("");
	wr.println("#include <" + s + ".h>");

	methods.printStubHeader(wr);
	globalVariables.printStubHeader(wr);

	wr.println("");
	wr.println("#endif /* !" + includeMacro + " */");
	wr.close();
    }

    /**
       ֤Υɤޤ
    */
    private void printStubCode() {
	String packageName = "stub_" + getPackageName();
	String name = getName();
	String dirName = packageName.replace('_', File.separatorChar);
	PrintWriter wr = createPrintWriter(dirName, name + ".c");

	String s = getPackageName() + "_" + getName();
	s = s.replace('_', File.separatorChar);

	wr.println("#include <stdlib.h>");
	wr.println("#include <stdarg.h>");
	wr.println("#include <" + s + ".h>");

	methods.printStubCode(wr);
	globalVariables.printStubCode(wr);
	wr.close();
    }

    /**
       Υ饹˽°륨ݡȤޤ

       @param elem export
       @return ݡ
    */
    public Export createExport(final Element elem) {
	Export m = new Export(this, elem);
	String type = m.getType();
	if (type.equals("struct")) {
	    structs.add(m);
	} else if (type.equals("union")) {
	    unions.add(m);
	} else if (type.equals("enum")) {
	    enums.add(m);
	} else {
	    throw new NodeException(elem, "unknown type");
	}
	addKeyword(m.getName());
	return m;
    }

    /**
       Υ饹˽°뷿ޤ

       @param elem typedef
       @return 
    */
    public Typedef createTypedef(final Element elem) {
	Typedef m = new Typedef(this, elem);
	addKeyword(m.getName());
	return m;
    }

    /**
       Υ饹˽°ޥޤ

       @param elem define
       @return ޥ
    */
    public Macro createMacro(final Element elem) {
	Macro m = new Macro(this, elem);
	addKeyword(m.getName());
	return m;
    }

    /**
       Υ饹˽°󷿤ޤ

       @param elem constant
       @return 󷿤
    */
    public Constant createConstant(final Element elem) {
	Constant m = new Constant(this, elem);
	addKeyword(m.getName());
	return m;
    }

    /**
       Υ饹˽°륢֥ȥ饯ȥ᥽åɤޤ

       @param elem abstractmethod
       @return ؿ
    */
    public Method createAbstractMethod(final Element elem) {
	Method m = new AbstractMethod(this, elem);
	addKeyword(m.getName());
	return m;
    }

    /**
       Υ饹˽°륰Хѿޤ

       @param elem globalvariable
       @return Хѿ
    */
    public GlobalVariable createGlobalVariable(final Element elem) {
	GlobalVariable m = new GlobalVariable(this, elem);
	addKeyword(m.getName());
	return m;
    }

    /**
       Υ饹˽°빽¤Τޤ϶ΤΥФޤ

       @param parentName Ф°빽¤Τޤ϶Τ̾
       @param elem memberǡstructmemberǡunionmemberǡ
       enummemberǡޤmethod
       @return ¤Τޤ϶ΤΥ
    */
    public Member createMember(final Strings parentName, final Element elem) {
	if (elem.getNodeName().equals("structmember")) {
	    Strings tag = expandImpliedNS(elem, "tag");
	    StructUnionMember m
		= new StructUnionMember(this, parentName, "struct", tag, elem);
	    structs.add(m);
	    addKeyword(m.getCompoundName());
	    return m;
	} else if (elem.getNodeName().equals("unionmember")) {
	    Strings tag = expandImpliedNS(elem, "tag");
	    StructUnionMember m
		= new StructUnionMember(this, parentName, "union", tag, elem);
	    unions.add(m);
	    addKeyword(m.getCompoundName());
	    return m;
	} else if (elem.getNodeName().equals("enummember")) {
	    Strings tag = expandImpliedNS(elem, "tag");
	    EnumMember m = new EnumMember(this, parentName, tag, elem);
	    enums.add(m);
	    addKeyword(m.getCompoundName());
	    return m;
	} else if (elem.getNodeName().equals("method")) {
	    Member m = new MethodMember(this, parentName, elem);
	    addKeyword(Strings.append(parentName, "." + m.getName()));
	    return m;
	} else if (elem.getNodeName().equals("member")) {
	    String type = expandAT(elem, "type");
	    importTags(type);
	    Member m = new Member(this, parentName, type, elem);
	    addKeyword(Strings.append(parentName, "." + m.getName()));
	    return m;
	} else {
	    throw new NodeException(elem, "unknown element name");
	}
    }

    /**
       Υ饹˽°빽¤Τޤ

       @param elem struct
       @return ¤
    */
    public StructUnion createStruct(final Element elem) {
	Strings tag = expandImpliedNS(elem, "tag");
	Strings as = expandImpliedNS(elem, "alias");

	if (tag.getQualified().equals("") && as.getQualified().equals("")) {
	    throw new NodeException(elem, "nameless struct.");
	}
	StructUnion m = new StructUnion(this, "struct", elem, tag, as);
	structs.add(m);
	addKeyword(m.getName());
	addKeyword(as);
	return m;
    }

    /**
       Υ饹˽°붦Τޤ

       @param elem union
       @return 
    */
    public StructUnion createUnion(final Element elem) {
	Strings tag = expandImpliedNS(elem, "tag");
	Strings as = expandImpliedNS(elem, "alias");

	if (tag.getQualified().equals("") && as.getQualified().equals("")) {
	    throw new NodeException(elem, "nameless union.");
	}
	StructUnion m = new StructUnion(this, "union", elem, tag, as);
	unions.add(m);
	addKeyword(m.getName());
	addKeyword(as);
	return m;
    }

    /**
       Υ饹˽°󷿤ޤ

       @param elem enum
       @return 
    */
    public Enum createEnum(final Element elem) {
	Strings tag = expandImpliedNS(elem, "tag");
	Strings as = expandImpliedNS(elem, "alias");

	if (tag.getQualified().equals("") && as.getQualified().equals("")) {
	    throw new NodeException(elem, "nameless enum.");
	}
	Enum m = new Enum(this, elem, tag, as);
	enums.add(m);
	addKeyword(m.getName());
	addKeyword(as);
	return m;
    }

    /**
       ̾Ÿ°ͤޤ

       @param elem 
       @param attrName °̾
       @return °ꤵƤʤжʸ󡢤Ǥʤ̾
       Ÿ°
    */
    private Strings expandImpliedNS(final Element elem,
				    final String attrName) {
	if (elem.getAttributeNode(attrName) == null) {
	    return Strings.intern("");
	}
	return expandNS(elem, attrName);
    }

    /**
       ݡȤ륿ɲäޤ

       @param s ʸ "(struct|union|enum) +(̾)"
       @return ɲäݡȤǤtrueˤfalse
    */
    public boolean addTag(final String s) {
	int k = s.indexOf(' ');
	String type = s.substring(0, k).trim();
	String tag = s.substring(k).trim();
	return tags.add(type + " " + tag);
    }

    /**
       {@inheritDoc}
    */
    @Override public void importTags(final String s) {
	String in = CommentedString.deleteComments(s);
	while (in.length() > 0) {
	    Matcher m = tagPattern.matcher(in);
	    if (!m.find()) {
		return;
	    }
	    int start = m.start();
	    int end = m.end();
	    String tag = in.substring(start, end);
	    addTag(tag);
	    in = in.substring(end);
	}
    }
}
